@testset "Transforms" begin
  @testset "Rotate" begin
    @test TB.isrevertible(Rotate)
    @test TB.isinvertible(Rotate)
    @test TB.inverse(Rotate(Angle2d(T(π / 2)))) == Rotate(Angle2d(-T(π / 2)))

    # ------
    # POINT
    # ------

    f = Rotate(Angle2d(T(π / 2)))
    g = P2(1, 0)
    r, c = TB.apply(f, g)
    @test r ≈ P2(0, 1)
    @test TB.revert(f, r, c) ≈ g

    # --------
    # SEGMENT
    # --------

    f = Rotate(Angle2d(T(π / 2)))
    g = Segment(P2(0, 0), P2(1, 0))
    r, c = TB.apply(f, g)
    @test r ≈ Segment(P2(0, 0), P2(0, 1))
    @test TB.revert(f, r, c) ≈ g

    # ----
    # BOX
    # ----

    f = Rotate(Angle2d(T(π / 2)))
    g = Box(P2(0, 0), P2(1, 1))
    r, c = TB.apply(f, g)
    @test r isa Quadrangle
    @test r ≈ Quadrangle(P2(0, 0), P2(0, 1), P2(-1, 1), P2(-1, 0))
    q = TB.revert(f, r, c)
    @test q isa Quadrangle
    @test q ≈ convert(Quadrangle, g)

    f = Rotate(V3(1, 0, 0), V3(0, 1, 0))
    g = Box(P3(0, 0, 0), P3(1, 1, 1))
    r, c = TB.apply(f, g)
    @test r isa Hexahedron
    @test r ≈ Hexahedron(
      P3(0, 0, 0),
      P3(0, 1, 0),
      P3(-1, 1, 0),
      P3(-1, 0, 0),
      P3(0, 0, 1),
      P3(0, 1, 1),
      P3(-1, 1, 1),
      P3(-1, 0, 1)
    )
    h = TB.revert(f, r, c)
    @test h isa Hexahedron
    @test h ≈ convert(Hexahedron, g)

    # ----------
    # ROPE/RING
    # ----------

    f = Rotate(Angle2d(T(π / 2)))
    g = Rope(P2(0, 0), P2(1, 0), P2(1, 1), P2(0, 1))
    r, c = TB.apply(f, g)
    @test r ≈ Rope(P2(0, 0), P2(0, 1), P2(-1, 1), P2(-1, 0))
    @test TB.revert(f, r, c) ≈ g

    f = Rotate(Angle2d(T(π / 2)))
    g = Ring(P2(0, 0), P2(1, 0), P2(1, 1), P2(0, 1))
    r, c = TB.apply(f, g)
    @test r ≈ Ring(P2(0, 0), P2(0, 1), P2(-1, 1), P2(-1, 0))
    @test TB.revert(f, r, c) ≈ g

    # ---------
    # TRIANGLE
    # ---------

    f = Rotate(AngleAxis(T(π / 2), T(0), T(0), T(1)))
    g = Triangle(P3(0, 0, 0), P3(1, 0, 0), P3(0, 1, 0))
    r, c = TB.apply(f, g)
    @test r ≈ Triangle(P3(0, 0, 0), P3(0, 1, 0), P3(-1, 0, 0))
    @test TB.revert(f, r, c) ≈ g

    f = Rotate(V3(0, 0, 1), V3(1, 0, 0))
    g = Triangle(P3(0, 0, 0), P3(1, 0, 0), P3(0, 1, 0))
    r, c = TB.apply(f, g)
    @test r ≈ Triangle(P3(0, 0, 0), P3(0, 0, -1), P3(0, 1, 0))
    @test TB.revert(f, r, c) ≈ g

    # ---------
    # POLYAREA
    # ---------

    f = Rotate(Angle2d(T(π / 2)))
    p = PolyArea(P2(0, 0), P2(1, 0), P2(1, 1), P2(0, 1))
    r, c = TB.apply(f, p)
    @test r ≈ PolyArea(P2(0, 0), P2(0, 1), P2(-1, 1), P2(-1, 0))
    @test TB.revert(f, r, c) ≈ p

    # ----------
    # MULTIGEOM
    # ----------

    f = Rotate(Angle2d(T(π / 2)))
    t = Triangle(P2(0, 0), P2(1, 0), P2(1, 1))
    g = Multi([t, t])
    r, c = TB.apply(f, g)
    @test r ≈ Multi([f(t), f(t)])
    @test TB.revert(f, r, c) ≈ g

    # ------
    # PLANE
    # ------

    f = Rotate(V3(0, 0, 1), V3(1, 0, 0))
    g = Plane(P3(0, 0, 0), V3(0, 0, 1))
    r, c = TB.apply(f, g)
    @test r ≈ Plane(P3(0, 0, 0), V3(1, 0, 0))
    @test TB.revert(f, r, c) ≈ g

    # ---------
    # CYLINDER
    # ---------

    f = Rotate(V3(0, 0, 1), V3(1, 0, 0))
    g = Cylinder(T(1))
    r, c = TB.apply(f, g)
    @test r ≈ Cylinder(P3(0, 0, 0), P3(1, 0, 0))
    @test TB.revert(f, r, c) ≈ g

    # ---------
    # POINTSET
    # ---------

    f = Rotate(Angle2d(T(π / 2)))
    d = PointSet([P2(0, 0), P2(1, 0), P2(1, 1)])
    r, c = TB.apply(f, d)
    @test r ≈ PointSet([P2(0, 0), P2(0, 1), P2(-1, 1)])
    @test TB.revert(f, r, c) ≈ d

    # ------------
    # GEOMETRYSET
    # ------------

    f = Rotate(Angle2d(T(π / 2)))
    t = Triangle(P2(0, 0), P2(1, 0), P2(1, 1))
    d = GeometrySet([t, t])
    r, c = TB.apply(f, d)
    @test r ≈ GeometrySet([f(t), f(t)])
    @test TB.revert(f, r, c) ≈ d
    d = [t, t]
    r, c = TB.apply(f, d)
    @test all(r .≈ [f(t), f(t)])
    @test all(TB.revert(f, r, c) .≈ d)

    # --------------
    # CARTESIANGRID
    # --------------

    f = Rotate(Angle2d(T(π / 2)))
    d = CartesianGrid{T}(10, 10)
    r, c = TB.apply(f, d)
    @test r ≈ SimpleMesh(f.(vertices(d)), topology(d))
    @test TB.revert(f, r, c) ≈ d

    # -----------
    # SIMPLEMESH
    # -----------

    f = Rotate(Angle2d(T(π / 2)))
    p = P2[(0, 0), (1, 0), (0, 1), (1, 1), (0.5, 0.5)]
    c = connect.([(1, 2, 5), (2, 4, 5), (4, 3, 5), (3, 1, 5)], Triangle)
    d = SimpleMesh(p, c)
    r, c = TB.apply(f, d)
    @test r ≈ SimpleMesh(f.(vertices(d)), topology(d))
    @test TB.revert(f, r, c) ≈ d
  end

  @testset "Translate" begin
    @test TB.isrevertible(Translate)
    @test TB.isinvertible(Translate)
    @test TB.inverse(Translate(T(1), T(2))) == Translate(T(-1), T(-2))

    # ------
    # POINT
    # ------

    f = Translate(T(1), T(1))
    g = P2(1, 0)
    r, c = TB.apply(f, g)
    @test r ≈ P2(2, 1)
    @test TB.revert(f, r, c) ≈ g

    # --------
    # SEGMENT
    # --------

    f = Translate(T(1), T(1))
    g = Segment(P2(0, 0), P2(1, 0))
    r, c = TB.apply(f, g)
    @test r ≈ Segment(P2(1, 1), P2(2, 1))
    @test TB.revert(f, r, c) ≈ g

    # ----
    # BOX
    # ----

    f = Translate(T(1), T(1))
    g = Box(P2(0, 0), P2(1, 1))
    r, c = TB.apply(f, g)
    @test r isa Box
    @test r ≈ Box(P2(1, 1), P2(2, 2))
    @test TB.revert(f, r, c) ≈ g

    # ---------
    # TRIANGLE
    # ---------

    f = Translate(T(1), T(2), T(3))
    g = Triangle(P3(0, 0, 0), P3(1, 0, 0), P3(0, 1, 1))
    r, c = TB.apply(f, g)
    @test r ≈ Triangle(P3(1, 2, 3), P3(2, 2, 3), P3(1, 3, 4))
    @test TB.revert(f, r, c) ≈ g

    # ----------
    # MULTIGEOM
    # ----------

    f = Translate(T(1), T(1))
    t = Triangle(P2(0, 0), P2(1, 0), P2(1, 1))
    g = Multi([t, t])
    r, c = TB.apply(f, g)
    @test r ≈ Multi([f(t), f(t)])
    @test TB.revert(f, r, c) ≈ g

    # ------
    # PLANE
    # ------

    f = Translate(T(0), T(0), T(1))
    g = Plane(P3(0, 0, 0), V3(0, 0, 1))
    r, c = TB.apply(f, g)
    @test r ≈ Plane(P3(0, 0, 1), V3(0, 0, 1))
    @test TB.revert(f, r, c) ≈ g

    # ---------
    # CYLINDER
    # ---------

    f = Translate(T(0), T(0), T(1))
    g = Cylinder(T(1))
    r, c = TB.apply(f, g)
    @test r ≈ Cylinder(P3(0, 0, 1), P3(0, 0, 2))
    @test TB.revert(f, r, c) ≈ g

    # ---------
    # POINTSET
    # ---------

    f = Translate(T(1), T(1))
    d = PointSet([P2(0, 0), P2(1, 0), P2(1, 1)])
    r, c = TB.apply(f, d)
    @test r ≈ PointSet([P2(1, 1), P2(2, 1), P2(2, 2)])
    @test TB.revert(f, r, c) ≈ d

    # ------------
    # GEOMETRYSET
    # ------------

    f = Translate(T(1), T(1))
    t = Triangle(P2(0, 0), P2(1, 0), P2(1, 1))
    d = GeometrySet([t, t])
    r, c = TB.apply(f, d)
    @test r ≈ GeometrySet([f(t), f(t)])
    @test TB.revert(f, r, c) ≈ d
    d = [t, t]
    r, c = TB.apply(f, d)
    @test all(r .≈ [f(t), f(t)])
    @test all(TB.revert(f, r, c) .≈ d)

    # --------------
    # CARTESIANGRID
    # --------------

    f = Translate(T(1), T(1))
    d = CartesianGrid{T}(10, 10)
    r, c = TB.apply(f, d)
    @test r isa CartesianGrid
    @test r ≈ CartesianGrid(P2(1, 1), P2(11, 11), dims=(10, 10))
    @test TB.revert(f, r, c) ≈ d

    # -----------
    # SIMPLEMESH
    # -----------

    f = Translate(T(1), T(1))
    p = P2[(0, 0), (1, 0), (0, 1), (1, 1), (0.5, 0.5)]
    c = connect.([(1, 2, 5), (2, 4, 5), (4, 3, 5), (3, 1, 5)], Triangle)
    d = SimpleMesh(p, c)
    r, c = TB.apply(f, d)
    @test r ≈ SimpleMesh(f.(vertices(d)), topology(d))
    @test TB.revert(f, r, c) ≈ d
  end

  @testset "Stretch" begin
    @test TB.isrevertible(Stretch)
    @test TB.isinvertible(Stretch)
    @test TB.inverse(Stretch(T(1), T(2))) == Stretch(T(1), T(1 / 2))

    # ------
    # POINT
    # ------

    f = Stretch(T(1), T(2))
    g = P2(1, 1)
    r, c = TB.apply(f, g)
    @test r ≈ P2(1, 2)
    @test TB.revert(f, r, c) ≈ g

    # --------
    # SEGMENT
    # --------

    f = Stretch(T(1), T(2))
    g = Segment(P2(0, 0), P2(1, 0))
    r, c = TB.apply(f, g)
    @test r ≈ Segment(P2(0, 0), P2(1, 0))
    @test TB.revert(f, r, c) ≈ g

    f = Stretch(T(2), T(1))
    g = Segment(P2(0, 0), P2(1, 0))
    r, c = TB.apply(f, g)
    @test r ≈ Segment(P2(0, 0), P2(2, 0))
    @test TB.revert(f, r, c) ≈ g

    # ----
    # BOX
    # ----

    f = Stretch(T(1), T(2))
    g = Box(P2(0, 0), P2(1, 1))
    r, c = TB.apply(f, g)
    @test r isa Box
    @test r ≈ Box(P2(0, 0), P2(1, 2))
    @test TB.revert(f, r, c) ≈ g

    # ---------
    # TRIANGLE
    # ---------

    f = Stretch(T(1), T(2), T(3))
    g = Triangle(P3(0, 0, 0), P3(1, 0, 0), P3(0, 1, 1))
    r, c = TB.apply(f, g)
    @test r ≈ Triangle(P3(0, 0, 0), P3(1, 0, 0), P3(0, 2, 3))
    @test TB.revert(f, r, c) ≈ g

    # ----------
    # MULTIGEOM
    # ----------

    f = Stretch(T(1), T(2))
    t = Triangle(P2(0, 0), P2(1, 0), P2(1, 1))
    g = Multi([t, t])
    r, c = TB.apply(f, g)
    @test r ≈ Multi([f(t), f(t)])
    @test TB.revert(f, r, c) ≈ g

    # ------
    # PLANE
    # ------

    f = Stretch(T(1), T(1), T(2))
    g = Plane(P3(1, 1, 1), V3(0, 0, 1))
    r, c = TB.apply(f, g)
    @test r ≈ Plane(P3(1, 1, 2), V3(0, 0, 1))
    @test TB.revert(f, r, c) ≈ g

    f = Stretch(T(2), T(1), T(1))
    g = Plane(P3(1, 1, 1), V3(0, 0, 1))
    r, c = TB.apply(f, g)
    @test r ≈ g
    @test TB.revert(f, r, c) ≈ g

    # ---------
    # CYLINDER
    # ---------

    f = Stretch(T(1), T(1), T(2))
    g = Cylinder(T(1))
    r, c = TB.apply(f, g)
    @test r ≈ Cylinder(P3(0, 0, 0), P3(0, 0, 2))
    @test TB.revert(f, r, c) ≈ g

    # ---------
    # POINTSET
    # ---------

    f = Stretch(T(1), T(2))
    d = PointSet([P2(0, 0), P2(1, 0), P2(1, 1)])
    r, c = TB.apply(f, d)
    @test r ≈ PointSet([P2(0, 0), P2(1, 0), P2(1, 2)])
    @test TB.revert(f, r, c) ≈ d

    # ------------
    # GEOMETRYSET
    # ------------

    f = Stretch(T(1), T(2))
    t = Triangle(P2(0, 0), P2(1, 0), P2(1, 1))
    d = GeometrySet([t, t])
    r, c = TB.apply(f, d)
    @test r ≈ GeometrySet([f(t), f(t)])
    @test TB.revert(f, r, c) ≈ d
    d = [t, t]
    r, c = TB.apply(f, d)
    @test all(r .≈ [f(t), f(t)])
    @test all(TB.revert(f, r, c) .≈ d)

    # --------------
    # CARTESIANGRID
    # --------------

    f = Stretch(T(1), T(2))
    d = CartesianGrid(P2(1, 1), P2(11, 11), dims=(10, 10))
    r, c = TB.apply(f, d)
    @test r isa CartesianGrid
    @test r ≈ CartesianGrid(P2(1, 2), P2(11, 22), dims=(10, 10))
    @test TB.revert(f, r, c) ≈ d

    # -----------
    # SIMPLEMESH
    # -----------

    f = Stretch(T(1), T(2))
    p = P2[(0, 0), (1, 0), (0, 1), (1, 1), (0.5, 0.5)]
    c = connect.([(1, 2, 5), (2, 4, 5), (4, 3, 5), (3, 1, 5)], Triangle)
    d = SimpleMesh(p, c)
    r, c = TB.apply(f, d)
    @test r ≈ SimpleMesh(f.(vertices(d)), topology(d))
    @test TB.revert(f, r, c) ≈ d
  end

  @testset "Expand" begin
    @test TB.isrevertible(Expand)
    @test TB.isinvertible(Expand)
    @test TB.inverse(Expand(T(1), T(2))) == Expand(T(1), T(1 / 2))

    # ------
    # POINT
    # ------

    f = Expand(T(1), T(2))
    g = P2(1, 1)
    r, c = TB.apply(f, g)
    @test r ≈ P2(1, 1)
    @test TB.revert(f, r, c) ≈ g

    # --------
    # SEGMENT
    # --------

    f = Expand(T(1), T(2))
    g = Segment(P2(0, 0), P2(1, 0))
    r, c = TB.apply(f, g)
    @test r ≈ Segment(P2(0, 0), P2(1, 0))
    @test TB.revert(f, r, c) ≈ g

    f = Expand(T(2), T(1))
    g = Segment(P2(0, 0), P2(1, 0))
    r, c = TB.apply(f, g)
    @test r ≈ Segment(P2(-0.5, 0), P2(1.5, 0))
    @test TB.revert(f, r, c) ≈ g

    # ----
    # BOX
    # ----

    f = Expand(T(1), T(2))
    g = Box(P2(0, 0), P2(1, 1))
    r, c = TB.apply(f, g)
    @test r isa Box
    @test r ≈ Box(P2(0, -0.5), P2(1, 1.5))
    @test TB.revert(f, r, c) ≈ g

    # ---------
    # TRIANGLE
    # ---------

    f = Expand(T(1), T(2), T(2))
    g = Triangle(P3(0, 0, 0), P3(1, 0, 0), P3(0, 1, 1))
    r, c = TB.apply(f, g)
    @test r ≈ Triangle(P3(0, -1 / 3, -1 / 3), P3(1, -1 / 3, -1 / 3), P3(0, 10 / 6, 10 / 6))
    @test TB.revert(f, r, c) ≈ g

    # ----------
    # MULTIGEOM
    # ----------

    f = Expand(T(1), T(2))
    t = Triangle(P2(0, 0), P2(1, 0), P2(1, 1))
    g = Multi([t, t])
    r, c = TB.apply(f, g)
    @test r ≈ Multi([f(t), f(t)])
    @test TB.revert(f, r, c) ≈ g

    # ------
    # PLANE
    # ------

    f = Expand(T(1), T(1), T(2))
    g = Plane(P3(1, 1, 1), V3(0, 0, 1))
    r, c = TB.apply(f, g)
    @test r ≈ g
    @test TB.revert(f, r, c) ≈ g

    f = Expand(T(2), T(1), T(1))
    g = Plane(P3(1, 1, 1), V3(0, 0, 1))
    r, c = TB.apply(f, g)
    @test r ≈ g
    @test TB.revert(f, r, c) ≈ g

    # ---------
    # CYLINDER
    # ---------

    f = Expand(T(1), T(1), T(2))
    g = Cylinder(T(1))
    r, c = TB.apply(f, g)
    @test r ≈ Cylinder(P3(0, 0, -0.5), P3(0, 0, 1.5))
    @test TB.revert(f, r, c) ≈ g

    # ---------
    # POINTSET
    # ---------

    f = Expand(T(1), T(2))
    d = PointSet([P2(0, 0), P2(1, 0), P2(1, 1)])
    r, c = TB.apply(f, d)
    @test r ≈ PointSet([P2(0, -1 / 3), P2(1, -1 / 3), P2(1, 10 / 6)])
    @test TB.revert(f, r, c) ≈ d

    # ------------
    # GEOMETRYSET
    # ------------

    f = Expand(T(1), T(2))
    t = Triangle(P2(0, 0), P2(1, 0), P2(1, 1))
    d = GeometrySet([t, t])
    r, c = TB.apply(f, d)
    @test r ≈ GeometrySet([f(t), f(t)])
    @test TB.revert(f, r, c) ≈ d
    d = [t, t]
    r, c = TB.apply(f, d)
    @test all(r .≈ [f(t), f(t)])
    @test all(TB.revert(f, r, c) .≈ d)

    # --------------
    # CARTESIANGRID
    # --------------

    f = Expand(T(1), T(2))
    d = CartesianGrid(P2(1, 1), P2(11, 11), dims=(10, 10))
    r, c = TB.apply(f, d)
    @test r isa CartesianGrid
    @test r ≈ CartesianGrid(P2(1, -4), P2(11, 16), dims=(10, 10))
    @test TB.revert(f, r, c) ≈ d

    # -----------
    # SIMPLEMESH
    # -----------

    f = Expand(T(1), T(2))
    p = P2[(0, 0), (1, 0), (0, 1), (1, 1), (0.5, 0.5)]
    c = connect.([(1, 2, 5), (2, 4, 5), (4, 3, 5), (3, 1, 5)], Triangle)
    d = SimpleMesh(p, c)
    r, c = TB.apply(f, d)
    @test r ≈ SimpleMesh(f(vertices(d)), topology(d))
    @test TB.revert(f, r, c) ≈ d
  end

  @testset "StdCoords" begin
    @test TB.isrevertible(StdCoords)

    # ---------
    # POINTSET
    # ---------

    f = StdCoords()
    d = view(PointSet(rand(P2, 100)), 1:50)
    r, c = TB.apply(f, d)
    @test all(sides(boundingbox(r)) .≤ T(1))
    @test TB.revert(f, r, c) ≈ d
    r2 = TB.reapply(f, d, c)
    @test r == r2

    # --------------
    # CARTESIANGRID
    # --------------

    f = StdCoords()
    d = CartesianGrid(P2(1, 1), P2(11, 11), dims=(10, 10))
    r, c = TB.apply(f, d)
    @test r isa CartesianGrid
    @test r ≈ CartesianGrid(P2(-0.5, -0.5), P2(0.5, 0.5), dims=(10, 10))
    @test TB.revert(f, r, c) ≈ d

    f = StdCoords()
    d = CartesianGrid{T}(10, 20)
    r, c = TB.apply(f, d)
    @test r ≈ CartesianGrid(P2(-0.5, -0.5), P2(0.5, 0.5), dims=(10, 20))
    @test TB.revert(f, r, c) ≈ d
    r2 = TB.reapply(f, d, c)
    @test r == r2
  end

  @testset "Repair{0}" begin
    poly = PolyArea(P2[(0, 0), (1, 0), (1, 0), (1, 1), (0, 1), (0, 1)])
    rpoly = poly |> Repair{0}()
    @test nvertices(rpoly) == 4
    @test vertices(rpoly) == P2[(0, 0), (1, 0), (1, 1), (0, 1)]
  end

  @testset "Repair{1}" begin
    # a tetrahedron with an unused vertex
    points = P3[(0, 0, 0), (0, 0, 1), (5, 5, 5), (0, 1, 0), (1, 0, 0)]
    connec = connect.([(1, 2, 4), (1, 2, 5), (1, 4, 5), (2, 4, 5)])
    mesh = SimpleMesh(points, connec)
    rmesh = mesh |> Repair{1}()
    @test nvertices(rmesh) == nvertices(mesh) - 1
    @test nelements(rmesh) == nelements(mesh)
    @test P3(5, 5, 5) ∉ vertices(rmesh)
  end

  @testset "Repair{2}" begin end

  @testset "Repair{3}" begin end

  @testset "Repair{4}" begin end

  @testset "Repair{5}" begin end

  @testset "Repair{6}" begin end

  @testset "Repair{7}" begin
    # mesh with inconsistent orientation
    points = rand(P3, 6)
    connec = connect.([(1, 2, 3), (3, 4, 2), (4, 3, 5), (6, 3, 1)])
    mesh = SimpleMesh(points, connec)
    rmesh = mesh |> Repair{7}()
    topo = topology(mesh)
    rtopo = topology(rmesh)
    e = collect(elements(topo))
    n = collect(elements(rtopo))
    @test n[1] == e[1]
    @test n[2] != e[2]
    @test n[3] != e[3]
    @test n[4] != e[4]
  end

  @testset "Repair{8}" begin
    poly =
      PolyArea(P2[(0.0, 0.0), (0.5, -0.5), (1.0, 0.0), (1.5, 0.5), (1.0, 1.0), (0.5, 1.5), (0.0, 1.0), (-0.5, 0.5)])
    rpoly = poly |> Repair{8}()
    @test nvertices(rpoly) == 4
    @test vertices(rpoly) == P2[(0.5, -0.5), (1.5, 0.5), (0.5, 1.5), (-0.5, 0.5)]

    # degenerate triangle with repeated vertices
    poly = PolyArea(P2[(0, 0), (1, 1), (1, 1)])
    rpoly = poly |> Repair{8}()
    @test !hasholes(rpoly)
    @test rings(rpoly) == [Ring(P2(0, 0))]
    @test vertices(rpoly) == [P2(0, 0)]
  end

  @testset "Repair{9}" begin
    poly = Quadrangle(P3(0, 1, 0), P3(1, 1, 0), P3(1, 0, 0), P3(0, 0, 0))
    bpoly = poly |> Repair{9}()
    @test bpoly isa Quadrangle
    @test bpoly == poly
  end

  @testset "Repair{10}" begin
    outer = Ring(P2[(0, 0), (0, 3), (2, 3), (2, 2), (3, 2), (3, 0)])
    inner = Ring(P2[(1, 1), (1, 2), (2, 2), (2, 1)])
    poly = PolyArea(outer, inner)
    repair = Repair{10}()
    rpoly, cache = TB.apply(repair, poly)
    @test nvertices(rpoly) == nvertices(poly)
    @test length(first(rings(rpoly))) > length(first(rings(poly)))
    opoly = TB.revert(repair, rpoly, cache)
    @test opoly == poly
  end

  @testset "Bridge" begin
    # https://github.com/JuliaGeometry/Meshes.jl/issues/566
    outer = Ring(P2(6, 4), P2(6, 7), P2(1, 6), P2(1, 1), P2(5, 2))
    inner₁ = Ring(P2(3, 3), P2(3, 4), P2(4, 3))
    inner₂ = Ring(P2(2, 5), P2(2, 6), P2(3, 5))
    poly = PolyArea([outer, inner₁, inner₂])
    bpoly = poly |> Bridge(T(0.1))
    @test !hasholes(bpoly)
    @test nvertices(bpoly) == 15

    # unique and bridges
    poly = PolyArea(P2[(0, 0), (1, 0), (1, 0), (1, 1), (1, 2), (0, 2), (0, 1), (0, 1)])
    cpoly = poly |> Repair{0}() |> Bridge()
    @test cpoly == PolyArea(P2[(0, 0), (1, 0), (1, 1), (1, 2), (0, 2), (0, 1)])

    # basic ngon tests
    t = Triangle(P2(0, 0), P2(1, 0), P2(0, 1))
    @test (t |> Bridge() |> boundary) == boundary(t)
    q = Quadrangle(P2(0, 0), P2(1, 0), P2(1, 1), P2(0, 1))
    @test (q |> Bridge() |> boundary) == boundary(q)

    # bridges between holes
    outer = P2[(0, 0), (1, 0), (1, 1), (0, 1)]
    hole1 = P2[(0.2, 0.2), (0.4, 0.2), (0.4, 0.4), (0.2, 0.4)]
    hole2 = P2[(0.6, 0.2), (0.8, 0.2), (0.8, 0.4), (0.6, 0.4)]
    poly = PolyArea([outer, hole1, hole2])
    @test vertices(poly) == P2[
      (0, 0),
      (1, 0),
      (1, 1),
      (0, 1),
      (0.2, 0.2),
      (0.2, 0.4),
      (0.4, 0.4),
      (0.4, 0.2),
      (0.6, 0.2),
      (0.6, 0.4),
      (0.8, 0.4),
      (0.8, 0.2)
    ]
    bpoly = poly |> Bridge(T(0.01))
    target = P2[
      (-0.0035355339059327372, 0.0035355339059327372),
      (0.19646446609406729, 0.20353553390593274),
      (0.2, 0.4),
      (0.4, 0.405),
      (0.6, 0.405),
      (0.8, 0.4),
      (0.8, 0.2),
      (0.6, 0.2),
      (0.6, 0.395),
      (0.4, 0.395),
      (0.4, 0.2),
      (0.20353553390593274, 0.19646446609406729),
      (0.0035355339059327372, -0.0035355339059327372),
      (1.0, 0.0),
      (1.0, 1.0),
      (0.0, 1.0)
    ]
    @test all(vertices(bpoly) .≈ target)

    poly = Quadrangle(P3(0, 1, 0), P3(1, 1, 0), P3(1, 0, 0), P3(0, 0, 0))
    bpoly = poly |> Bridge()
    @test bpoly isa Quadrangle
    @test bpoly == poly
  end

  @testset "Smoothing" begin
    # smoothing doesn't change the topology
    trans = LaplaceSmoothing(30)
    @test TB.isrevertible(trans)
    mesh = readply(T, joinpath(datadir, "beethoven.ply"))
    smesh = trans(mesh)
    @test nvertices(smesh) == nvertices(mesh)
    @test nelements(smesh) == nelements(mesh)
    @test topology(smesh) == topology(mesh)

    # smoothing doesn't change the topology
    trans = TaubinSmoothing(30)
    @test TB.isrevertible(trans)
    mesh = readply(T, joinpath(datadir, "beethoven.ply"))
    smesh = trans(mesh)
    @test nvertices(smesh) == nvertices(mesh)
    @test nelements(smesh) == nelements(mesh)
    @test topology(smesh) == topology(mesh)
  end
end
